#include "display.h"
#include <stdlib.h>
#include <stdio.h>

#include <xcb/render.h>
#include <xcb/xfixes.h>
#include <xcb/damage.h>

d_display_t*
display_open(const char *name)
{
    d_display_t *dpy = NULL;
    xcb_connection_t *conn;

    conn = xcb_connect(name, NULL);
    if (conn) {
        dpy = malloc(sizeof(d_display_t));
        dpy->conn = conn;
        dpy->ref = 1;
    }
    return dpy;
}

void
display_ref(d_display_t *dpy)
{
    ++dpy->ref;
}

void
display_unref(d_display_t *dpy)
{
    if (dpy && --dpy->ref == 0) {
        xcb_disconnect(dpy->conn);
        free(dpy);
    }
}

void
display_error(d_display_t *dpy, xcb_generic_error_t *ev)
{
    const int major_opcode = ((xcb_request_error_t*)ev)->major_opcode;
    const int minor_opcode = ((xcb_request_error_t*)ev)->minor_opcode;
    const char *name, *req;

    /* XXX check if we should ignore it (ev->full_sequence) */

    (void)dpy;

    name = NULL;
    switch (ev->error_code) {
    case XCB_REQUEST:   name = "BadRequest";   break;
    case XCB_VALUE:     name = "BadValue";     break;
    case XCB_WINDOW:    name = "BadWindow";    break;
    case XCB_PIXMAP:    name = "BadPixmap";    break;
    case XCB_MATCH:     name = "BadMatch";     break;
    case XCB_DRAWABLE:  name = "BadDrawable";  break;
    case XCB_G_CONTEXT: name = "BadGC";        break;
    default: break;
    }
    if (!name)
        switch (ev->error_code - dpy->fixes_error) {
        case XCB_XFIXES_BAD_REGION: name = "BadRegion"; break;
        default: break;
        }
    if (!name)
        switch (ev->error_code - dpy->damage_error) {
        case XCB_DAMAGE_BAD_DAMAGE: name = "BadDamage"; break;
        default: break;
        }
    if (!name)
        switch (ev->error_code - dpy->render_error) {
        case XCB_RENDER_PICT_FORMAT: name = "BadPictFormat"; break;
        case XCB_RENDER_PICT_OP:     name = "BadPictOp"; break;
        case XCB_RENDER_PICTURE:     name = "BadPicture"; break;
        case XCB_RENDER_GLYPH_SET:   name = "BadGlyphSet"; break;
        case XCB_RENDER_GLYPH:       name = "BadGlyph"; break;
        default: break;
        }

    if (major_opcode <= 127)
        switch (major_opcode) {
        case 1:  req = "CreateWindow";           break;
        case 2:  req = "ChangeWindowAttributes"; break;
        case 3:  req = "GetWindowAttributes";    break;
        case 14: req = "GetGeometry";            break;
        case 15: req = "QueryTree";              break;
        case 18: req = "ChangeProperty";         break;
        case 20: req = "GetProperty";            break;
        case 22: req = "SetSelectionOwner";      break;
        case 23: req = "GetSelectionOwner";      break;
        case 53: req = "CreatePixmap";           break;
        case 54: req = "FreePixmap";             break;
        case 55: req = "CreateGC";               break;
        case 56: req = "ChangeGC";               break;
        case 60: req = "FreeGC";                 break;
        case 72: req = "PutImage";               break;
        default: break; 
        }

    if (name && req)
        printf("XError: %s %s!\n",
               name, req);
    else if (name)
        printf("XError: %s major opcode %d minor opcode %d!\n",
               name, major_opcode, minor_opcode);
    else
        printf("XError: code %d major opcode %d minor opcode %d!\n",
               ev->error_code, major_opcode, minor_opcode);

    abort();
}
